package MooX::ValidateSubs::Role;

use Moo::Role;
use Carp qw/croak/;
use Type::Utils qw//;
use Type::Params qw/compile compile_named/;
use Types::Standard qw//;

sub _validate_sub {
	my ( $self, $name, $type, $spec, @params ) = @_;
	my $store_spec = sprintf '%s_spec', $name;

	my $compiled_check = ($self->$store_spec->{"compiled_$type"} ||= do {
		if (ref $spec eq 'ARRAY') {
			my @types = map {
				my ($constraint, $default) = (@$_, 0);
				$default eq '1' ? Types::Standard::Optional->of($constraint) : $constraint;
			} @$spec;
			compile(@types);
		}
		else {
			my %types;
			for my $key (keys %$spec) {
				my ($constraint, $default) = (@{$spec->{$key}}, 0);
				$types{$key} =
					$default eq '1' ? Types::Standard::Optional->of($constraint) : $constraint;
			}
			compile_named(%types);
		}
	});

	my @count = ( scalar @params );
	if ( ref $spec eq 'ARRAY' ) {
		push @count, scalar @{$spec};

		if ( $count[0] == 1 && $count[1] != 1 and ref $params[0] eq 'ARRAY' ) {
			@params = @{ $params[0] };
			$count[0] = scalar @params;
			$count[3] = 'ref';
		}

		$count[2] = $count[1] - grep { $spec->[$_]->[1] } 0 .. $count[1] - 1;
		$count[0] >= $count[2] && $count[0] <= $count[1]
			or croak sprintf 'Error - Invalid count in %s for sub - %s - expected - %s - got - %s',
				$type, $name, $count[1], $count[0];

		foreach ( 0 .. $count[1] - 1 ) {
			not $params[$_] and $spec->[$_]->[1]
				and ( $spec->[$_]->[1] =~ m/^1$/ and next or $params[$_] = $self->_default( $spec->[$_]->[1] ) );
		}

		@params = $compiled_check->(@params);
		return defined $count[3] ? \@params : @params;
	}

	my %para = $count[0] == 1 ? %{ $params[0] } : @params;
	my %cry = ( %{$spec}, %para );
	foreach ( keys %cry ) {
		not $para{$_} and $spec->{$_}->[1]
			and ( $spec->{$_}->[1] =~ m/^1$/ and next or $para{$_} = $self->_default( $spec->{$_}->[1] ) );
	}

	my $paraRef = $compiled_check->(\%para);

	return $count[0] == 1 ? $paraRef : %{$paraRef};
}

sub _default {
	my ( $self, $default ) = @_;

	if ( ref $default eq 'CODE' ) {
		return $default->();
	}
	return $self->$default;
}

sub _preprocess_params {
	my ($self, @params) = @_;

	my @world;
	for my $p ( @params ) {
		my ($ref, $index, $arg) = (ref $p);
		! $ref and do { ($index, $arg) = $p =~ m/(\d+)=(.*)/ } && (defined $index && defined $arg)
				? @world = _generate_params($arg, $index, @world)
				: push @world, $p
			or ( $ref eq 'HASH' ) and (map { $p->{$_} = [$self->_preprocess_params($p->{$_})]->[0] } keys %{$p})
				&& push @world, $p
			or ($ref eq 'ARRAY' ) and push @world, [ $self->_preprocess_params(@{$p}) ]
			or push @world, $p;
	}
	return @world;
}

sub _generate_params {
	my ($string, $index, @world, $current) = @_;

	( $_ =~ m/^\s*\[\s*(.*)\]/ )
		and do {
			$current = [_generate_params($1)];
			$world[$index] = $current;
		}
		or ( $_ =~ m/^\s*\{\s*(.*)\s*\}/ ) and do {
			$current = {};
			my %temp = split '=>', $1;
			$current->{[_generate_params(($_))]->[0] } = [_generate_params($temp{$_})]->[0]
				for keys %temp;
			$world[$index] = $current;
		} or (not defined $index) and do {
			(($_ =~ m/^\s*(\d+)\s*$/) || ($_ =~ m/^\s*[\'\"]+\s*(.*?)\s*[\'\"]+\s*$/))
				? push @world, $1 : do { $_ =~ s/^\s+|\s+$//g } && push @world, $_;
		} or $world[$index] = $string foreach ( split /(?![^\(\[\{]+[\]\}\)]),/, $string );

	return @world;
}



1;

